
#include <dolfin.h>
#include <fmg.h>
#include "Elasticity.h"

using namespace dolfin;

// Dirichlet boundary condition for clamp at left end
class Clamp : public Expression
{
public:
	Clamp() : Expression(3) {}

	void eval(Array<double>& values, const Array<double>& x) const
	{
		values[0] = 0.0;
		values[1] = 0.0;
		values[2] = 0.0;
	}

};

// Sub domain for clamp at left end
class Left : public SubDomain
{
	bool inside(const Array<double>& x, bool on_boundary) const
	{
		return x[0] < DOLFIN_EPS;
	}
};

class Load : public Expression
{
public:
	Load() : Expression(3) {}

	void eval(Array<double>& values, const Array<double>& x) const
	{
		double r2 = 4.0/5.0-DOLFIN_EPS;
		values[0] = 0.0;
		values[1] = (x[0] > r2) ? -10*exp(2*x[2]+x[1]) : 0.0;
		values[2] = 0.0;
	}

};

class ShapeDom : public SubDomain
{
public:
	bool inside(const Array<double>& x, bool on_boundary) const
	{
		double r1 = 1.0/5.0+DOLFIN_EPS;
		double r2 = 4.0/5.0-DOLFIN_EPS;
		return ((x[0] < r1) || (x[0] > r2) || (x[1] < r1) || (x[1] > r2)) &&
		       ((x[0] < r1) || (x[0] > r2) || (x[2] < r1) || (x[2] > r2)) &&
		       ((x[1] < r1) || (x[1] > r2) || (x[2] < r1) || (x[2] > r2));
	}
};

int main(int argc, char* argv[])
{
	SubMesh mesh(UnitCube(5,5,5), ShapeDom());

	// Read mesh and create function space
	Elasticity::FunctionSpace V(mesh);

	// Create right-hand side
	Load f;
	//Constant f(1,0);
	// Set up boundary condition at left end
	Clamp c;
	Left left;
	DirichletBC bcl(V, c, left);

	// Collect boundary conditions
	std::vector<const BoundaryCondition*> bcs;
	bcs.push_back(&bcl);

	// Set elasticity parameters
	double E = 210e3;
	double nu = 0.3;
	Constant mu(E / (2*(1 + nu)));
	Constant lambda(E*nu / ((1 + nu)*(1 - 2*nu)));

	// Define variational problem
	Elasticity::BilinearForm a(V, V);
	a.mu = mu; a.lmbda = lambda;
	Elasticity::LinearForm L(V);
	L.f = f;
	Function u(V);
	LinearVariationalProblem problem(a, L, u, bcs);

	fmg::Tests tests(problem);
	tests.parameters["num_refinements"] = 3;
	tests.parameters["min_coarse_depth"] = 2;
	tests.parameters.parse(argc, argv);
	tests.run();

	return 0;
}
