
#include <dolfin.h>
#include <fmg.h>
#include "MixedPoisson.h"

using namespace dolfin;

// Source term (right-hand side)
class Source : public Expression
{
  void eval(Array<double>& values, const Array<double>& x) const
  {
    double dx = x[0] - 0.5;
    double dy = x[1] - 0.5;
    values[0] = 10*exp(-(dx*dx + dy*dy) / 0.02);
  }
};

// Boundary source for flux boundary condition
class BoundarySource : public Expression
{
public:

  BoundarySource(const Mesh& mesh) : Expression(2), mesh(mesh) {}

  void eval(Array<double>& values, const Array<double>& x,
            const ufc::cell& ufc_cell) const
  {
    dolfin_assert(ufc_cell.local_facet >= 0);

    const Mesh* pmesh = &mesh;
    while (pmesh != NULL && (pmesh->id() != ufc_cell.mesh_identifier))
      pmesh = pmesh->child_shared_ptr().get();

    Cell cell(*pmesh, ufc_cell.index);
    Point n = cell.normal(ufc_cell.local_facet);

    const double g = sin(5*x[0]);
    values[0] = g*n[0];
    values[1] = g*n[1];
  }

private:

  const Mesh& mesh;

};

// Sub domain for essential boundary condition
class EssentialBoundary : public SubDomain
{
  bool inside(const Array<double>& x, bool on_boundary) const
  {
    return x[1] < DOLFIN_EPS or x[1] > 1.0 - DOLFIN_EPS;
  }
};

int main(int argc, char* argv[])
{
  // Create mesh
  UnitSquare mesh(4, 4);

  // Construct function space
  MixedPoisson::FunctionSpace W(mesh);
  MixedPoisson::BilinearForm a(W, W);
  MixedPoisson::LinearForm L(W);

  // Create source and assign to L
  Source f;
  L.f = f;

  // Define boundary condition
  SubSpace W0(W, 0);
  BoundarySource G(mesh);
  EssentialBoundary boundary;
  DirichletBC bc(W0, G, boundary);

  Function w(W);
  LinearVariationalProblem problem(a, L, w, bc);

  fmg::Tests tests(problem);
  tests.parameters["num_refinements"] = 4;
  tests.parameters["write_solutions"] = true;
  tests.parameters.parse(argc, argv);
  tests.run();
     
  return 0;
}
