
// Eddy currents phenomena in low conducting body can be
// described using electric vector potential and curl-curl operator:
//		\nabla \times \nabla \times T = - \frac{\partial B}{\partial t}
// Electric vector potential defined as:
//		\nabla \times T = J
//
// Boundary condition
//		J_n = 0,
//		T_t=T_w=0, \frac{\partial T_n}{\partial n} = 0
// which is naturaly fulfilled for zero Dirichlet BC with Nedelec (edge)
// elements.

#include <dolfin.h>
#include <fmg.h>
#include "EddyCurrents.h"
#include "CurrentDensity.h"

using namespace dolfin;


// Homogenous external magnetic field (dB/dt)
class Source : public Expression
{
public:

	Source() : Expression(3) {}

	void eval(Array<double>& values, const Array<double>& x) const
	{
		values[0] = 0.0;
		values[1] = 0.0;
		values[2] = 1.0;
	}

};


// Zero Dirichlet BC
class Zero : public Expression
{
public:

	Zero() : Expression(3) {}

	void eval(Array<double>& values, const Array<double>& x) const
	{
		values[0] = 0.0;
		values[1] = 0.0;
		values[2] = 0.0;
	}

};


// Everywhere on external surface
class DirichletBoundary: public SubDomain
{
	bool inside(const Array<double>& x, bool on_boundary) const
	{
		return on_boundary;
	}
};


int main()
{
	//fmg::Utils::DEBUG = true;

	// Create demo mesh
	UnitCube mesh(4,4,4);

	// Define functions
	Source dbdt;
	Zero zero;

	// Define function space and boundary condition
	EddyCurrents::FunctionSpace V(mesh);
	DirichletBoundary boundary;
	DirichletBC bc(V, zero, boundary);

	// Define variational problem for T
	EddyCurrents::BilinearForm a(V,V);
	EddyCurrents::LinearForm L(V);
	L.dbdt = dbdt;

	// Compute solution
	Function T(V);
	LinearVariationalProblem problem(a, L, T, bc);
	fmg::MultigridProblem mg_problem(problem);

	mg_problem.adapt();
	mg_problem.adapt();
	mg_problem.adapt();

	fmg::MultigridSolver mg_solver(mg_problem);
	//mg_solver.parameters["pre_smoother_reps"] = 2;
	//mg_solver.parameters["pre_smoother"] = "ssor";
	//mg_solver.parameters["post_smoother"] = "ssor";
	//mg_solver.parameters["post_smoother_reps"] = 2;

	fmg::MultigridPreconditionedKrylovSolver solver(mg_solver);
	solver.parameters["monitor_convergence"] = true;

	// There is some kind of locking going on, when measuring in the preconditioned norm.
	// (Linf-norm of solution x on the coarse level goes only to 1e-3 while rhs b actually goes to zero
	// and residual Linf-norm b-A*x is constant at about 1e-16 (since using a direct solver))
	// This seems to be not a bug in the solver, rather a numerical property of the problem
	// This is why we do not use preconditioned norm here.
#if FMG_DOLFIN_VERSION > 1
	solver.parameters["convergence_norm_type"] = "true";
#else
	KSPSetNormType(*solver.ksp(), KSP_NORM_UNPRECONDITIONED);
#endif

	solver.solve();

	Function& T_fine = *solver.solution();

	// Define variational problem for J
	CurrentDensity::FunctionSpace V1(mesh);
	CurrentDensity::BilinearForm a1(V1,V1);
	CurrentDensity::LinearForm L1(V1);
	CurrentDensity::CoefficientSpace_T VT(mesh);
	Function T1(VT);
	L1.T = T1;

	// Solve problem
	Function J(V1);
	LinearVariationalProblem problem1(a1, L1, J);
	fmg::MultigridProblem mg_problem1(problem1);

	for (dolfin::uint i = 1; i < mg_problem.num_levels(); i++) {
		mg_problem1.adapt(mg_problem.level(i).V->mesh());
	}

	fmg::MultigridSolver mg_solver1(mg_problem1);
	fmg::MultigridPreconditionedKrylovSolver solver1(mg_solver1);
	solver1.parameters["monitor_convergence"] = true;

	// Set form coefficient T
	Form* L_fine = const_cast<Form*>(solver1.mg_solver().finest_level().L.get());
	Function& T_coef = *((Function*)L_fine->coefficient(0).get());
	*(T_coef.vector()) = *(T_fine.vector());

	solver1.solve();

	Function& J_fine = *solver1.solution();

	File file("current_density.pvd");
	file << J_fine;

	return 0;
}
