
#include <dolfin.h>
#include <fmg.h>
#include "Stokes.h"
#include "StokesK.h"
#include "StokesM.h"

using namespace dolfin;

// Function for no-slip boundary condition for velocity
class Noslip : public Expression
{
public:
	Noslip() : Expression(2) {}

	void eval(Array<double>& values, const Array<double>& x) const
	{
		values[0] = 0.0;
		values[1] = 0.0;
	}
};

class NoslipDomain : public SubDomain
{
        bool inside(const Array<double>& x, bool on_boundary) const
        {
                return on_boundary && !(x[1] >= 1.0-DOLFIN_EPS);
        }
};

// Function for inflow boundary condition for velocity
class Inflow : public Expression
{
public:
	Inflow() : Expression(2) {}

	void eval(Array<double>& values, const Array<double>& x) const
	{
		values[0] = 1.0;
		values[1] = 0.0;
	}
};

class InflowDomain : public SubDomain
{
        bool inside(const Array<double>& x, bool on_boundary) const
        {
                return x[1] >= 1.0-DOLFIN_EPS;
        }
};


class StokesBlockPrec : public PETScUserPreconditioner
{
public:
	dolfin::uint Kiter;
	dolfin::uint Miter;
	dolfin::uint nSolve;

	StokesBlockPrec(dolfin::uint nK, dolfin::uint nM, PETScUserPreconditioner& K_prec, PETScUserPreconditioner& M_prec) :
		_nK(nK), _K_prec(reference_to_no_delete_pointer(K_prec)),
		_M_prec(reference_to_no_delete_pointer(M_prec)),
		_u(), _bu(), _p(), _bp()
	{
		_indices_K.resize(nK);
		for (dolfin::uint i = 0; i < nK; i++) {
			_indices_K[i] = i;
		}

		_indices_M.resize(nM);
		for (dolfin::uint i = 0; i < nM; i++) {
			_indices_M[i] = i+nK;
		}

		_u.reset(new PETScVector(nK));
		_bu.reset(new PETScVector(nK));
		_p.reset(new PETScVector(nM));
		_bp.reset(new PETScVector(nM));

		Kiter = Miter = nSolve = 0;
	}

	void solve(dolfin::PETScVector& x, const dolfin::PETScVector& b)
	{
#if 0
		x = b;
		return;
#endif
		PetscScalar* pb;
		PetscScalar* px;

		VecGetArray(*(b.vec()), &pb);
		VecGetArray(*(x.vec()), &px);
		VecPlaceArray(*(_bu->vec()), pb);
		VecPlaceArray(*(_bp->vec()), pb+_nK);
		VecPlaceArray(*(_u->vec()), px);
		VecPlaceArray(*(_p->vec()), px+_nK);

		{
			fmg::FMGTimer t_prec_K("t_prec_K");
			_K_prec->solve(*_u, *_bu);
			Kiter++;
		}
		{
			fmg::FMGTimer t_prec_K("t_prec_M");
			_M_prec->solve(*_p, *_bp);
			Miter++;
		}

		VecResetArray(*(_bu->vec()));
		VecResetArray(*(_bp->vec()));
		VecResetArray(*(_u->vec()));
		VecResetArray(*(_p->vec()));
		VecRestoreArray(*(b.vec()), &pb);
		VecRestoreArray(*(x.vec()), &px);

		nSolve++;
	}

protected:
	dolfin::uint _nK;
	boost::shared_ptr<PETScUserPreconditioner> _K_prec;
	boost::shared_ptr<PETScUserPreconditioner> _M_prec;
	boost::shared_ptr<PETScVector> _u;
	boost::shared_ptr<PETScVector> _bu;
	boost::shared_ptr<PETScVector> _p;
	boost::shared_ptr<PETScVector> _bp;
	std::vector<dolfin::uint> _indices_K;
	std::vector<dolfin::uint> _indices_M;
};


class ShapeDom : public SubDomain
{
public:
	static bool obstacle;

        bool inside(const Array<double>& x, bool on_boundary) const
        {
                return x[1] >= 0.3-DOLFIN_EPS && x[1] <= 0.7+DOLFIN_EPS &&
                       (obstacle ? !(x[0] > 0.4+DOLFIN_EPS && x[1] > 0.4+DOLFIN_EPS && x[0] < 0.6-DOLFIN_EPS && x[1] < 0.6-DOLFIN_EPS &&
			      (fabs(x[0]-x[1]) > 0.1+DOLFIN_EPS)) : true);
        }
};

bool ShapeDom::obstacle = true;


class LUPreconditioner : public dolfin::PETScUserPreconditioner
{
public:
	LUPreconditioner(const boost::shared_ptr<const PETScMatrix>& A) {
		_lu_solver.reset(new PETScLUSolver());
		_lu_solver->set_operator(A);
		_lu_solver->parameters["reuse_factorization"] = true;
		_lu_solver->parameters["same_nonzero_pattern"] = true;
		if (_lu_solver->parameters.has_parameter("symmetric_operator"))
			_lu_solver->parameters["symmetric_operator"] = true;
	}

	void solve(dolfin::PETScVector& x, const dolfin::PETScVector& b) {
		_lu_solver->solve(x,b );
	}

protected:
	boost::shared_ptr<PETScLUSolver> _lu_solver;
};


class SplittingPreconditioner : public dolfin::PETScUserPreconditioner
{
public:
	SplittingPreconditioner(const boost::shared_ptr<const PETScMatrix>& A, const std::string& method, double omega = 1.0) {
		_method.reset(new fmg::SmoothingOperator(A, method, omega));
	}

	void solve(dolfin::PETScVector& x, const dolfin::PETScVector& b) {
		bool zero = true;
		_method->apply(x, b, zero, 1);
	}

protected:
	boost::shared_ptr<fmg::SmoothingOperator> _method;
};


class OtherPreconditioner : public dolfin::PETScUserPreconditioner
{
public:
	OtherPreconditioner(const boost::shared_ptr<const PETScBaseMatrix>& A, const std::string& method) {
		_solver.reset(new PETScKrylovSolver("cg"));
		_solver->set_operator(A);
		/*_solver->parameters["error_on_nonconvergence"] = false;
		_solver->parameters["nonzero_initial_guess"] = false;
		_solver->parameters["relative_tolerance"] = 0.0;
		_solver->parameters["absolute_tolerance"] = 0.0;
		_solver->parameters["maximum_iterations"] = 2;*/
		_method.reset(new PETScPreconditioner(method));
		_method->set(*_solver);

		PC pc;
		KSPGetPC(*(_solver->ksp()), &pc);
		::PCSetOperators(pc, *(A->mat()), *(A->mat()), SAME_PRECONDITIONER);
	}


	void solve(dolfin::PETScVector& x, const dolfin::PETScVector& b) {
		PC pc;
		KSPGetPC(*(_solver->ksp()), &pc);
		::PCApply(pc, *(b.vec()), *(x.vec()));
		//_solver->solve(x, b);
	}

protected:
	boost::shared_ptr<PETScPreconditioner> _method;
	boost::shared_ptr<PETScKrylovSolver> _solver;
};



void run_test(fmg::Table& tab, std::string method, std::string prec,
	const Parameters& solver_params, const Parameters& mg_params, const Parameters& prec_params, dolfin::uint nrefine)
{
	fmg::FMGTimer t_init("init_time");

	fmg::FMGTimer::reset_all();

	// Create mesh
	UnitSquare mesh(2,2);

	// Create function space and subspaces
	Stokes::FunctionSpace W(mesh);

	Function w(W);
	Function u = w[0];
	Function p = w[1];
	const FunctionSpace& W0 = *u.function_space();
	const FunctionSpace& W1 = *p.function_space();

	// Velocity preconditioner
	// Create functions for boundary conditions
	Noslip noslip;
	Inflow inflow;
	Constant zero(0);

	// No-slip boundary condition for velocity
	NoslipDomain noslip_dom;
	DirichletBC bc0(W0, noslip, noslip_dom);

	// Inflow boundary condition for velocity
	InflowDomain inflow_dom;
	DirichletBC bc1(W0, inflow, inflow_dom);

	// Velocity preconditioner
	StokesK::BilinearForm Ka(W0, W0);
	StokesK::LinearForm KL(W0);
	KL.z = zero;
	std::vector<const BoundaryCondition*> bcs_u;
	bcs_u.push_back(&bc0); bcs_u.push_back(&bc1);
	LinearVariationalProblem KP(Ka, KL, u, bcs_u);
	fmg::MultigridSolver mgs_K(KP, nrefine);
	mgs_K.parameters.update(mg_params);
	mgs_K.init();

	// Pressure preconditioner
	StokesM::BilinearForm Ma(W1, W1);
	StokesM::LinearForm ML(W1);
	ML.z = zero;
	LinearVariationalProblem MP(Ma, ML, p);
	fmg::MultigridSolver mgs_M(MP, nrefine);
	mgs_M.parameters.update(mg_params);
	mgs_M.init();

	// Define fine problem

	// Define variational problem
	const Mesh& meshf = W0.mesh()->leaf_node();
	const FunctionSpace& Wf = adapt(W, reference_to_no_delete_pointer(meshf));
	FunctionSpace& W0f = *Wf[0];
	FunctionSpace& W1f = *Wf[1];

	// Collect boundary conditions
	std::vector<const DirichletBC*> bcs;
	DirichletBC bc0f(W0f, noslip, noslip_dom);
	DirichletBC bc1f(W0f, inflow, inflow_dom);
	bcs.push_back(&bc0f);
	bcs.push_back(&bc1f);

	Constant f(0.0, 0.0);
	Stokes::BilinearForm a(Wf, Wf);
	Stokes::LinearForm L(Wf);
	L.f = f;

	PETScMatrix A;
	PETScVector b;
	assemble_system(A, b, a, L, bcs);

#if 0
	fmg::Utils::print_matrix("A", A);
	fmg::Utils::print_matrix("K", *mgs_K.finest_level().A);
	fmg::Utils::print_matrix("M", *mgs_M.finest_level().A);
	return 0;
#endif

	boost::shared_ptr<const PETScMatrix> K = reference_to_no_delete_pointer(*((const PETScMatrix*) mgs_K.finest_level().A.get()));
	boost::shared_ptr<const PETScMatrix> M = reference_to_no_delete_pointer(*((const PETScMatrix*) mgs_M.finest_level().A.get()));
	dolfin::uint nK = K->size(0);
	dolfin::uint nM = M->size(0);

	boost::shared_ptr<PETScUserPreconditioner> prec_K;
	boost::shared_ptr<PETScUserPreconditioner> prec_M;

	//prec_M.reset(new LUPreconditioner(M));
	prec_M.reset(new SplittingPreconditioner(M, "jacobi"));

	if (prec == "fmg") {
		prec_K.reset(new fmg::MultigridPreconditioner(mgs_K));
	}
	else if (prec == "lu") {
		prec_K.reset(new LUPreconditioner(K));
	}
	else {
		prec_K.reset(new OtherPreconditioner(K, prec));
	}

	//prec_K->parameters.update(prec_params);

	StokesBlockPrec P(nK, nM, *prec_K, *prec_M);
	PETScKrylovSolver solver(method, P);

	solver.parameters.update(solver_params);

   //solver.parameters["monitor_convergence"] = true;
	if (solver.parameters.has_parameter("convergence_norm_type"))
		solver.parameters["convergence_norm_type"] = "preconditioned";
	else
		KSPSetNormType(*(solver.ksp()), KSP_NORM_PRECONDITIONED);

	tab["t_init"] << t_init.stop();
	tab["nrefine"] << nrefine;
	tab["dofs"] << (nK+nM);
	tab["K_dofs"] << nK;
	tab["M_dofs"] << nM;

	// Solve
	fmg::FMGTimer t_solve("t_solve");
	Function wf(Wf);
	dolfin::uint iter = solver.solve(A, *(wf.vector()), b);

	tab["t_solve"] << t_solve.stop();
	tab["t_prec_K"] << fmg::FMGTimer::get_sum("t_prec_K");
	tab["t_prec_M"] << fmg::FMGTimer::get_sum("t_prec_M");
//	tab["iter_prec_K"] << P.Kiter;
//	tab["iter_prec_M"] << P.Miter;
//	tab["iter_prec_K_avg"] << (P.Kiter/(double)iter);
//	tab["iter_prec_M_avg"] << (P.Miter/(double)iter);
	tab["iter"] << iter;

	Function uf = wf[0];
	Function pf = wf[1];
	uf.rename("v", "velocity");
	pf.rename("p", "pressure");

	// Save solution in VTK format
	std::stringstream s;
	s << "velocity" << nrefine << ".pvd";
	File ufile_pvd(s.str());
	ufile_pvd << uf;
	s.str("");
	s << "pressure" << nrefine << ".pvd";
	File pfile_pvd(s.str());
	pfile_pvd << pf;

	tab.finish_row();
}


int main(int argc, char* argv[])
{
	// disable dof reordering to preserve block structure of matrices
        if (parameters.has_key("reorder_dofs_serial"))
                parameters["reorder_dofs_serial"] = false;
        else
                parameters["reorder_dofs"] = false;

	dolfin::Parameters p(fmg::Tests::default_parameters());
	dolfin::Parameters pp(PETScPreconditioner::default_parameters());
	dolfin::Parameters sp(PETScKrylovSolver::default_parameters());

	if (sp.has_parameter("convergence_norm_type"))
		sp["convergence_norm_type"] = "preconditioned";

	pp.rename("prec_params");
	sp.rename("solver_params");

	p["primary_solver"] = "minres";
	p["num_refinements"] = 7;
	p.add("obstacle", true);
	p.add("prec", "fmg");
	p.add(pp);
	p.add(sp);
	p.parse(argc, argv);

	ShapeDom::obstacle = p["obstacle"];
	dolfin::uint nrefine = p["num_refinements"];
	std::string method = p["primary_solver"];
	std::string prec = p["prec"];
	Parameters mg_params = p("fmg_solver");
	Parameters prec_params = p("prec_params");
	Parameters solver_params = p("solver_params");
	fmg::Table tab(fmg::Table::parse_format(p["table_format"]));

	tab.add_int_column("nrefine", "nrefine");
	tab.add_int_column("dofs", "dofs", 6);
	tab.add_int_column("K_dofs", "K dofs", 6);
	tab.add_int_column("M_dofs", "M dofs", 5);
	tab.add_float_column("t_init", "init time");
//	tab.add_int_column("iter_prec_K", "prec_K iter");
//	tab.add_float_column("iter_prec_K_avg", "(avg)");
	tab.add_float_column("t_prec_K", "prec_K time");
//	tab.add_int_column("iter_prec_M", "prec_M iter");
//	tab.add_float_column("iter_prec_M_avg", "(avg)");
	tab.add_float_column("t_prec_M", "prec_M time");
	tab.add_int_column("iter", method + " iter");
	tab.add_float_column("t_solve", method + " time");
	tab.finish_header();

	dolfin::LogManager::logger.set_log_active(false);

	for (dolfin::uint n = 0; n <= nrefine; n++) {
		run_test(tab, method, prec, solver_params, mg_params, prec_params, n);
	}

	return 0;
}
