/*
 * ProlongationAssembler.cpp
 *
 *  Created on: 02.10.2012
 *      Author: Felix Ospald
 */

#include "ProlongationAssembler.h"
#include "../common/common.h"

#include <dolfin/function/FunctionSpace.h>
#include <dolfin/fem/GenericDofMap.h>
#include <dolfin/la/GenericVector.h>
#include <dolfin/la/PETScMatrix.h>

#include <limits>

#if FMG_DOLFIN_VERSION > 1
	#include <dolfin/la/TensorLayout.h>
#endif

namespace fmg {


// This class is used to pass a basis function to the evaluate_dof function
class basis_func : public ufc::function {
	public:
		const dolfin::FiniteElement* element;
		const ufc::cell* cell;
		uint ibasis;

		inline void evaluate(double* values, const double* x, const ufc::cell& c) const
	    {
	    	element->evaluate_basis(ibasis, values, x, *cell);
	    }
};


#define USE_SPACE_GROUPING

// This is the central method for building the prolongation matrix P.
boost::shared_ptr<dolfin::GenericMatrix> ProlongationAssembler::assemble_matrix(
		const dolfin::FunctionSpace& space,
		const dolfin::FunctionSpace& refined_space,
		const dolfin::GenericLinearAlgebraFactory& factory,
		AssemblyMethod asm_method,
		DirichletProjectionMethod dbc_method,
		BoostFlag boost_flags,
		const std::vector<int>* global_to_reduced_dofs,
		const std::vector<uint>* reduced_to_global_dofs,
		const std::vector<int>* refined_global_to_reduced_dofs,
		const std::vector<uint>* refined_reduced_to_global_dofs,
		double eps)
{
	// assumptions:
	// - space.element() == refined_space.element()
	// - refined mesh contains same vertices (i.e. only (recursive) subdivison of cells allowed)

	bool space_grouping = ((boost_flags & BoostFlags::space_grouping) != 0);
	bool hashing = ((boost_flags & BoostFlags::hashing) != 0);
	uint rdim = refined_space.dim();
	const dolfin::Mesh& mesh = *(space.mesh());
	const dolfin::Mesh& rmesh = *(refined_space.mesh());
	std::vector<boost::shared_ptr<const dolfin::FunctionSpace> > subspaces;
	std::vector<boost::shared_ptr<const dolfin::FunctionSpace> > rsubspaces;
	std::vector<bool> computed_rdofs(rdim);
	dolfin::CellIterator cellit(mesh);
	dolfin::Cell* cell = &(*cellit);
	dolfin::UFCCell ufc_rcell(rmesh);
	dolfin::UFCCell ufc_cell(*cell);
	std::vector<uint> ops_rindex;
	std::vector<uint> ops_index;
	std::vector<double> ops_factor;
	std::vector<uint> nnz(rdim);
	//dof_eval_hash eval_hash;
	basis_func basis;
	uint cell_ratio = 0;
	double value;

	// get list of scalar subspaces
	Utils::extract_subspaces_recursive(dolfin::reference_to_no_delete_pointer(space), subspaces);
	Utils::extract_subspaces_recursive(dolfin::reference_to_no_delete_pointer(refined_space), rsubspaces);
	uint subspaces_size = subspaces.size();

	// estimate maximum required number of operations for prolongation
	uint max_ops = 0;
	for (uint i = 0; i < rsubspaces.size(); i++) {
		boost::shared_ptr<const dolfin::FiniteElement> relement = rsubspaces[i]->element();
		// FunctionSpace::dim() = number of all dofs in function space
		// FiniteElement::space_dimension() = number of dofs on each cell
		max_ops += rsubspaces[i]->dim()*relement->space_dimension();
	}

	// improve performance by grouping spaces with the same element signature
	std::vector<uint> space_group_size(rsubspaces.size(), 1);
#ifdef USE_SPACE_GROUPING
	if (space_grouping) {
		uint iref = 0;
		for (uint i = 1; i < rsubspaces.size(); i++) {
			if (rsubspaces[i]->element()->signature() == rsubspaces[iref]->element()->signature()) {
				space_group_size[iref] ++;
				continue;
			}
			iref = i;
		}
	}
#endif

	// init vectors
	uint nops = 0;
	ops_rindex.resize(max_ops);
	ops_index.resize(max_ops);
	ops_factor.resize(max_ops);

	// assembly method specific init
	switch (asm_method)
	{
		case AssemblyMethods::uniform:
			cell_ratio = rmesh.num_cells()/mesh.num_cells();
			if (cell_ratio*mesh.num_cells() != rmesh.num_cells()) {
				dolfin::dolfin_error(__FILE__, "assemble_matrix",
					"Uniform assembly method does not allow a rational ratio between cell numbers!");
			}
			break;
		case AssemblyMethods::cellwise:
			break;
		case AssemblyMethods::dofwise:
		default:
			dolfin::dolfin_error(__FILE__, "assemble_matrix",
				"Specified assembly method not implemented!");
			break;
	}

	// iterate over cells of refined mesh
	for (dolfin::CellIterator rcell(rmesh); !rcell.end(); ++rcell)
	{
		ufc_rcell.update(*rcell);

		if (asm_method == AssemblyMethods::uniform)
		{
			if ((rcell->index() % cell_ratio) == 0) {
				cell = &(*cellit);
				ufc_cell.update(*cell);
				++cellit;
			}
		}
		else if (asm_method == AssemblyMethods::cellwise)
		{
			// get cell of coarse mesh which contains refined cell
			// TODO: replace midpoint() and intersects(...) by something cheaper
			dolfin::Point rcellmidpoint = rcell->midpoint();

			if (!cell->intersects(rcellmidpoint))
			{
				// first see if next cell contains refined cell
				++cellit;

				if (!cellit.end() && cellit->intersects(rcellmidpoint)) {
					cell = &(*cellit);
				}
				else {
					// this is an relative expensive operation
					int icell = mesh.intersected_cell(rcellmidpoint);

					if (icell < 0) {
						// cell not found (mesh domains do not match?)
						dolfin::dolfin_error(__FILE__, "assemble_matrix",
								"Error matching cells. Probably the mesh and the refined mesh have a different domain?");
					}

					cell = &(cellit[icell]);
				}

				ufc_cell.update(*cell);
			}
		}

		basis.cell = &ufc_cell;

		// compute prolongation for each subspace
		for (size_t isub = 0; isub < subspaces_size; isub += space_group_size[isub])
		{
			const dolfin::FunctionSpace& subspace = *(subspaces[isub]);
			const dolfin::FunctionSpace& rsubspace = *(rsubspaces[isub]);
			const dolfin::FiniteElement& element = *(subspace.element());
			const dolfin::FiniteElement& relement = *(rsubspace.element());
			const dolfin::GenericDofMap& dofmap = *(subspace.dofmap());
			const dolfin::GenericDofMap& rdofmap = *(rsubspace.dofmap());
			const std::vector<uint>& dofs = dofmap.cell_dofs(cell->index());
			const std::vector<uint>& rdofs = rdofmap.cell_dofs(rcell->index());
			uint dofs_size = dofs.size();
			uint rdofs_size = rdofs.size();

			basis.element = &element;

			// iterate over all dof coordinates
			for (size_t irdof = 0; irdof < rdofs_size; irdof++)
			{
				// check if already computed
				if (computed_rdofs[rdofs[irdof]]) continue;

				// iterate over dofs of cell
				for (size_t idof = 0; idof < dofs_size; idof++)
				{
					// evaluate dof
					basis.ibasis = idof;

					//if (hashing) {
						/*std::cout << rcell->orientation() << " " << cell->orientation() << "\n";
						Utils::print_mesh_entities(*cell,0);
						std::cout << rcell->orientation() << " " << cell->orientation() << "\n";
						Utils::print_mesh_entities(*rcell,0);*/ //2*(int)cell->orientation()-1, 2*(int)rcell->orientation()-1
						// this does not work yet
					//	value = eval_hash.evaluate_dof(isub, relement, irdof, basis, ufc_rcell,
					//		ufc_rcell.geometric_dimension, rcell->num_entities(0), 1, 1);
					//}
					//else {
						value = relement.evaluate_dof(irdof, basis, ufc_rcell);
					//}

					if (fabs(value) > eps)
					{
#ifdef USE_SPACE_GROUPING
						for (size_t k = 0; k < space_group_size[isub]; k++)
						{
							// add operation
							ops_rindex[nops] = rsubspaces[isub+k]->dofmap()->cell_dofs(rcell->index())[irdof];
							ops_index[nops] = subspaces[isub+k]->dofmap()->cell_dofs(cell->index())[idof];
							ops_factor[nops] = value;
							// count nonzeros for each row
							nnz[ops_rindex[nops]]++;
							nops++;
						}
#else
						// add operation
						ops_rindex[nops] = rdofs[irdof];
						ops_index[nops] = dofs[idof];
						ops_factor[nops] = value;
						nops++;
						// count nonzeros for each row
						nnz[rdofs[irdof]]++;
#endif
					}
				}

				computed_rdofs[rdofs[irdof]] = true;
			}
		}
	}

	//if (hashing && Utils::DEBUG) {
	//	std::cout << "fmg: eval hash hits: " << eval_hash._hits << " allocs: " << eval_hash._allocs << " compares: " << eval_hash._compares << "\n";
	//}

	// create matrix representation of operator

	std::pair<uint, uint> ij;
	boost::shared_ptr<dolfin::GenericMatrix> P;
	boost::shared_ptr<Mat> PMat(new Mat);

	switch (dbc_method)
	{
	case DirichletProjectionMethods::zero:
	case DirichletProjectionMethods::remove:

		if (refined_global_to_reduced_dofs == NULL || global_to_reduced_dofs == NULL)
			dolfin::error("The mapping parameters are required!");

		// adjust nnz (saves some memory)
		for (uint i = 0; i < nops; i++) {
			if ((*refined_global_to_reduced_dofs)[ops_rindex[i]] < 0) {
				nnz[ops_rindex[i]] = 0;
				continue;
			}
			if (nnz[ops_rindex[i]] == 0)
				continue;
			if ((*global_to_reduced_dofs)[ops_index[i]] < 0)
				nnz[ops_rindex[i]] --;
		}
		break;

	case DirichletProjectionMethods::keep:
		break;
	}

	dolfin::SubSystemsManager::init_petsc();

	switch (dbc_method)
	{
	case DirichletProjectionMethods::zero:

		MatCreateSeqAIJ(PETSC_COMM_SELF, refined_space.dim(), space.dim(), 0,
				reinterpret_cast<PetscInt*>(&(nnz[0])), PMat.get());
		P.reset(new dolfin::PETScMatrix(PMat));

		for (uint i = 0; i < nops; i++) {
			if ((*refined_global_to_reduced_dofs)[ops_rindex[i]] < 0 || (*global_to_reduced_dofs)[ops_index[i]] < 0)
				continue;
			ij.first = ops_rindex[i];
			ij.second = ops_index[i];
			P->setitem(ij, ops_factor[i]);
		}
		break;

	case DirichletProjectionMethods::remove:
		{
			if (refined_reduced_to_global_dofs == NULL || reduced_to_global_dofs == NULL)
				dolfin::error("The mapping parameters are required!");

			// re-map nnz
			std::vector<uint> rnnz(refined_reduced_to_global_dofs->size());
			for (uint i = 0; i < rdim; i++) {
				if ((*refined_global_to_reduced_dofs)[i] < 0) continue;
				rnnz[(*refined_global_to_reduced_dofs)[i]] = nnz[i];
			}

			MatCreateSeqAIJ(PETSC_COMM_SELF, refined_reduced_to_global_dofs->size(), reduced_to_global_dofs->size(), 0,
					reinterpret_cast<PetscInt*>(&(rnnz[0])), PMat.get());
			P.reset(new dolfin::PETScMatrix(PMat));
		}

		for (uint i = 0; i < nops; i++) {
			if ((*refined_global_to_reduced_dofs)[ops_rindex[i]] < 0 || (*global_to_reduced_dofs)[ops_index[i]] < 0)
				continue;
			ij.first = (*refined_global_to_reduced_dofs)[ops_rindex[i]];
			ij.second = (*global_to_reduced_dofs)[ops_index[i]];
			P->setitem(ij, ops_factor[i]);
		}
		break;

	case DirichletProjectionMethods::keep:

		MatCreateSeqAIJ(PETSC_COMM_SELF, refined_space.dim(), space.dim(), 0,
				reinterpret_cast<PetscInt*>(&(nnz[0])), PMat.get());
		P.reset(new dolfin::PETScMatrix(PMat));

		for (uint i = 0; i < nops; i++) {
			ij.first = ops_rindex[i];
			ij.second = ops_index[i];
			P->setitem(ij, ops_factor[i]);
		}
		break;
	}

	P->apply("insert");
	return P;
}

/* This stuff is currently not used. It was inteded to speed up dof evaluation, but does currently not work.

class cell_hash
{
public:
	cell_hash(const cell_hash& ch) : _coords(ch._coords), _hash(ch._hash), _index(ch._index), _rel_index(ch._rel_index) { }

	typedef float round_type;

	round_type round(double v)
	{
		unsigned int w = (unsigned int) (((*((unsigned long *) &v)) >> 36) << 4);
		return *((round_type*) &w);
	}

	cell_hash(const ufc::cell& cell, uint geo_dim, uint coord_dim, const ufc::cell& rel_cell)
	{
		_coords.reset(new std::vector<round_type>(geo_dim*coord_dim));
		init(cell, geo_dim, coord_dim, rel_cell);
	}

	void init(const ufc::cell& cell, uint geo_dim, uint coord_dim, const ufc::cell& rel_cell)
	{
		_index = cell.index;
		_rel_index = rel_cell.index;
		_hash = 0;

		std::vector<round_type>& c = *_coords;
		uint k = 0;

		for (uint i = 0; i < coord_dim; i++) {
			for (uint j = 0; j < geo_dim; j++) {
				c[k] = round(cell.coordinates[i][j] - rel_cell.coordinates[0][j]);
				_hash += (k+1)*boost::hash_value(c[k]);
				k++;
			}
		}
	}

	void update(const ufc::cell& cell, uint geo_dim, uint coord_dim, const ufc::cell& rel_cell)
	{
	//	if (cell.index == _index && rel_cell.index == _rel_index) return;
		init(cell, geo_dim, coord_dim, rel_cell);
	}

	boost::shared_ptr<std::vector<round_type> > _coords;
	std::size_t _hash;
	std::size_t _index;
	std::size_t _rel_index;
};


bool operator==(const cell_hash& a, const cell_hash& b);
std::size_t hash_value(const cell_hash& ch);


class dof_eval_hash
{
public:
	// boost::unordered_map
	typedef boost::unordered_map<cell_hash, double> map_rcell;
	typedef boost::unordered_map<cell_hash, boost::shared_ptr<map_rcell> > map_cell;
	typedef boost::unordered_map<uint, boost::shared_ptr<map_cell> > map_rdof;
	typedef boost::unordered_map<uint, boost::shared_ptr<map_rdof> > map_dof;
	typedef boost::unordered_map<uint, boost::shared_ptr<map_dof> > map_sub;

	uint _hits;
	uint _allocs;
	static uint _compares;

	dof_eval_hash() : _hits(0), _allocs(0), has_sub(false), has_dof(false), has_rdof(false), has_cell(false), has_rcell(false) { _compares = 0; }

	map_sub::iterator sub;
	bool has_sub;
	map_dof::iterator dof;
	bool has_dof;
	map_rdof::iterator rdof;
	bool has_rdof;
	map_cell::iterator cell;
	bool has_cell;
	map_rcell::iterator rcell;
	bool has_rcell;
	boost::shared_ptr<cell_hash> _cell_hash;
	boost::shared_ptr<cell_hash> _rcell_hash;

	inline double evaluate_dof(uint isub, const dolfin::FiniteElement& relement, uint irdof, const basis_func& basis, const ufc::cell& ufc_rcell, uint geo_dim, uint coord_dim,
		int cell_orientation, int rcell_orientation)
	{
		if (!_cell_hash)
			_cell_hash.reset(new cell_hash(*(basis.cell), geo_dim, coord_dim, *(basis.cell)));
		else
			_cell_hash->update(*(basis.cell), geo_dim, coord_dim, *(basis.cell));

		if (!_rcell_hash)
			_rcell_hash.reset(new cell_hash(ufc_rcell, geo_dim, coord_dim, *(basis.cell)));
		else
			_rcell_hash->update(ufc_rcell, geo_dim, coord_dim, *(basis.cell));

		double value = relement.evaluate_dof(irdof, basis, ufc_rcell);

	//	if (has_sub) has_sub = sub->first == isub;
	//	if (!has_sub) {
			sub = _map_sub.find(isub);
			has_sub = sub != _map_sub.end();
	//	}
		if (has_sub) {
		//	if (has_dof) has_dof = dof->first == cell_orientation*basis.ibasis;
		//	if (!has_dof) {
				dof = sub->second->find(cell_orientation*basis.ibasis);
				has_dof = dof != sub->second->end();
		//	}
			if (has_dof) {
			//	if (has_rdof) has_rdof = rdof->first == rcell_orientation*irdof;
			//	if (!has_rdof) {
					rdof = dof->second->find(rcell_orientation*irdof);
					has_rdof = rdof != dof->second->end();
			//	}
				if (has_rdof) {
				//	if (has_cell) has_cell = cell->first == *_cell_hash;
				//	if (!has_cell) {
						cell = rdof->second->find(*_cell_hash);
						has_cell = cell != rdof->second->end();
				//	}
					if (has_cell) {
					//	if (has_rcell) has_rcell = rcell->first == *_rcell_hash;
					//	if (!has_rcell) {
							rcell = cell->second->find(*_rcell_hash);
							has_rcell = rcell != cell->second->end();
					//	}
						if (has_rcell) {
							_hits++;

							if (fabs(rcell->second-value) > 1e-11) {
								std::cout << "mismatch " << fabs(rcell->second-value) << "\n";
							}

							return rcell->second;
						}
					}
					else {
						has_rcell = false;
					}
				}
				else {
					has_cell = has_rcell = false;
				}
			}
			else {
				has_rdof = has_cell = has_rcell = false;
			}
		}
		else {
			has_dof = has_rdof = has_cell = has_rcell = false;
		}

		if (!has_sub) {
			boost::shared_ptr<map_dof> _map_dof(new map_dof());
			sub = _map_sub.insert(std::make_pair(isub, _map_dof)).first;
			has_sub = true;
		}
		if (!has_dof) {
			boost::shared_ptr<map_rdof> _map_rdof(new map_rdof());
			dof = sub->second->insert(std::make_pair(cell_orientation*basis.ibasis, _map_rdof)).first;
			has_dof = true;
		}
		if (!has_rdof) {
			boost::shared_ptr<map_cell> _map_cell(new map_cell());
			rdof = dof->second->insert(std::make_pair(rcell_orientation*irdof, _map_cell)).first;
			has_rdof = true;
		}
		if (!has_cell) {
			boost::shared_ptr<map_rcell> _map_rcell(new map_rcell());
			cell = rdof->second->insert(std::make_pair(*_cell_hash, _map_rcell)).first;
			has_cell = true;
		}
		if (!has_rcell) {
			rcell = cell->second->insert(std::make_pair(*_rcell_hash, value)).first;
			has_rcell = true;
		}

		_allocs++;
		return value;
	}

protected:
	map_sub _map_sub;
};


uint dof_eval_hash::_compares = 0;


bool operator==(const cell_hash& a, const cell_hash& b)
{
	if ((a._hash == b._hash) && (a._coords->size() == b._coords->size())) {
		dof_eval_hash::_compares++;
		return (memcmp(&((*a._coords)[0]), &((*b._coords)[0]), sizeof(cell_hash::round_type)*a._coords->size()) == 0);
		//return (*a._coords) == (*b._coords);
	}

	return false;
}


std::size_t hash_value(const cell_hash& ch)
{
    return ch._hash;
}

 */
} // namespace
