/*
 * Table.cpp
 *
 *  Created on: 01.10.2012
 *      Author: Felix Ospald
 */

#include "Table.h"

#include <iostream>
#include <iomanip>
#include <string.h>
#include <boost/format.hpp>

#include <dolfin/log/log.h>

namespace fmg {

// This could be done better, using an abstract class for each TableFormat,
// but is sufficient for now.


std::stringstream& Table::Column::get_stream()
{
	value << std::setw(width) << std::setprecision(precision) << std::showpoint;

	if (align > 0)
		value << std::right;
	else
		value << std::left;

	return value;
}


void Table::add_text_column(std::string key, std::string title, std::size_t width)
{
	add_column(key, title, width, 0, -1);
}


void Table::add_float_column(std::string key, std::string title, std::size_t precision)
{
	std::size_t width = std::max(precision+5, title.size());
	add_column(key, title, width, precision, 1);
}


void Table::add_int_column(std::string key, std::string title, std::size_t width)
{
	add_column(key, title, width, 0, 1);
}


void Table::add_column(std::string key, std::string title, std::size_t width, std::size_t precision, int align)
{
	if (_column_indices.count(key) > 0) {
		dolfin::error("Column key already exists!");
	}

	if (width == 0) {
		width = title.size();
	}

	width = std::max(width, title.size());

	switch (_format) {
		case TableFormats::tab:
		case TableFormats::csv:
			width = 0;
			align = 0;
			break;
		case TableFormats::console:
		case TableFormats::latex:
		case TableFormats::rest:
			break;
	}

	_column_indices[key] = _columns.size();
	_columns.push_back(boost::shared_ptr<Table::Column>(new Table::Column(width, precision, align)));
	_width += width;
	(*this)[key] << title;
}


void Table::finish_header()
{
	std::cout << std::endl;

	switch (_format) {
		case TableFormats::console:
			finish_row(); add_delimiter();
			break;
		case TableFormats::tab:
		case TableFormats::csv:
			finish_row();
			break;
		case TableFormats::latex:
			std::cout << "\\begin{tabular}{|";
			for (std::vector<boost::shared_ptr<Column> >::iterator i = _columns.begin(); i != _columns.end(); i++) {
				std::cout << "l|";
			}
			std::cout << "}" << std::endl;
			add_delimiter(); finish_row(); add_delimiter();
			break;
		case TableFormats::rest:
			add_delimiter('-', '+'); finish_row(); add_delimiter('=','+');
			break;
	}
}


void Table::finish_row()
{
	const char* firstcol = "";
	const char* colsep = "";
	const char* lastcol = "";
	bool resize = true;

	switch (_format) {
		case TableFormats::console: colsep = " | "; break;
		case TableFormats::csv: colsep = ","; resize = false; break;
		case TableFormats::tab: colsep = ","; resize = false; break;
		case TableFormats::latex: colsep = " & "; lastcol = "\\\\\\hline"; resize = false; break;
		case TableFormats::rest: firstcol = "| "; colsep = " | "; lastcol = " |"; break;
	}

	std::cout << firstcol;
	for (std::vector<boost::shared_ptr<Column> >::iterator i = _columns.begin(); i != _columns.end(); i++) {
		if (i != _columns.begin()) std::cout << colsep;
		std::string value = escape((*i)->value.str());
		if (resize) value.resize((*i)->width, ' ');
		std::cout << value;
		(*i)->value.str("");
	}
	std::cout << lastcol << std::endl;

	switch (_format) {
		case TableFormats::console: break;
		case TableFormats::csv: break;
		case TableFormats::tab: break;
		case TableFormats::latex: break;
		case TableFormats::rest: if (_rows > 0) add_delimiter('-', '+'); break;
	}

	_rows++;
}


void Table::add_delimiter()
{
	switch (_format) {
		case TableFormats::console: add_delimiter('-', '+'); break;
		case TableFormats::csv: break;
		case TableFormats::tab: break;
		case TableFormats::latex: std::cout << "\\hline" << std::endl; break;
		case TableFormats::rest: break;
	}
}


// Adds a horizontal delimiter line. hchar denotes the character for the horizontal line, and xchar
// the character at horizontal/vertical line crossings.
void Table::add_delimiter(char hchar, char xchar)
{
	switch (_format) {
		case TableFormats::console: break;
		case TableFormats::csv: break;
		case TableFormats::tab: break;
		case TableFormats::latex: break;
		case TableFormats::rest: std::cout << xchar << hchar;  break;
	}

	for (std::vector<boost::shared_ptr<Column> >::iterator i = _columns.begin(); i != _columns.end(); i++) {
		if (i != _columns.begin()) std::cout << hchar << xchar << hchar;
		std::cout << std::string((*i)->width, hchar);
	}

	switch (_format) {
		case TableFormats::console: break;
		case TableFormats::csv: break;
		case TableFormats::tab: break;
		case TableFormats::latex: break;
		case TableFormats::rest: std::cout << hchar << xchar;  break;
	}

	std::cout << std::endl;
}


// Escape special characters in string.
std::string Table::escape(const std::string& s)
{
	const char* c = s.c_str();
	const char* delim = "\t";
	std::string es;

	switch (_format) {
		case TableFormats::console: es = s; break;
		case TableFormats::csv:
			delim = ",";
		case TableFormats::tab:
			while (*c != '\0') {
				if (strchr(delim, *c) != NULL) { es.append(1,'\\'); es.append(1,*c); }
				else es.append(1,*c);
				c++;
			}
			break;
		case TableFormats::latex:
			while (*c != '\0') {
				if (strchr("_%{}$", *c) != NULL) { es.append(1,'\\'); es.append(1,*c); }
				else if (*c == '"') es.append(2,'`');
				else es.append(1,*c);
				c++;
			}
			break;
		case TableFormats::rest: es = s;  break;
	}

	return es;
}


Table::~Table()
{
	switch (_format) {
		case TableFormats::console: break;
		case TableFormats::csv: break;
		case TableFormats::tab: break;
		case TableFormats::latex: std::cout << "\\end{tabular}" << std::endl; break;
		case TableFormats::rest: break;
	}
}


Table::TableFormat Table::parse_format(const std::string& s)
{
	Table::TableFormat table_format = Table::TableFormats::console;

	if (s == "console")
		table_format = Table::TableFormats::console;
	else if (s == "csv")
		table_format = Table::TableFormats::csv;
	else if (s == "tab")
		table_format = Table::TableFormats::tab;
	else if (s == "latex")
		table_format = Table::TableFormats::latex;
	else if (s == "rest")
		table_format = Table::TableFormats::rest;
	else
		dolfin::error("Unknown table format '%s'", s.c_str());

	return table_format;
}


} // namespace
