/*
 * Table.h
 *
 *  Created on: 01.11.2012
 *      Author: Felix Ospald
 */

#ifndef FMG_TABLE_H_
#define FMG_TABLE_H_

#include <vector>
#include <map>
#include <sstream>
#include <boost/smart_ptr/shared_ptr.hpp>

namespace fmg {

/// This class is used to pretty print tables.
/// Example usage to print a 2x2 Table to std::cout::
///
///   Table tab;
///
///   tab.add_text_column("name", "Name", 6);
///   tab.add_float_column("time", "Time");
///   tab.finish_header();
///
///   tab["name"] << "x";
///   tab["time"] << 1.1;
///   tab.finish_row();
///
///   tab["name"] << "y";
///   tab["time"] << 1.1;
///   tab.finish_row();
///
/// .. note:: The *<<* operator should be used only once for each column until calling finish_row.

class Table
{
public:
	struct Column
	{
		Column(std::size_t w, std::size_t p, int a) : width(w), precision(p), align(a) { }
		std::stringstream& get_stream();
		std::size_t width;
		std::size_t precision;
		std::stringstream value;
		int align;
	}; // class

	typedef struct { enum e { console, csv, tab, latex, rest }; } TableFormats;
	typedef TableFormats::e TableFormat;

	Table(TableFormat format = TableFormats::console) : _format(format), _width(0), _rows(0) { }
	~Table();

	/// Adds a text column to the table definition.
	void add_text_column(std::string key, std::string title, std::size_t width = 0);

	/// Adds a float column to the table definition.
	void add_float_column(std::string key, std::string title, std::size_t precision = 3);

	/// Adds a integer column to the table definition.
	void add_int_column(std::string key, std::string title, std::size_t width = 0);

	/// Prints the table header. This should be called after all *add_..._column* calls.
	void finish_header();

	/// Prints the current table row. This should be called after setting the
	/// column values for the current row using the [] operator.
	void finish_row();

	/// Adds a row delimiter used for grouping several rows together.
	void add_delimiter();

	/// Access the stringstream of the column *key* for the current row.
	std::stringstream& operator[](std::string key) { return _columns[_column_indices[key]]->get_stream(); }

	/// Convert table format name to enum value.
	static TableFormat parse_format(const std::string& s);

protected:
	void add_column(std::string key, std::string title, std::size_t width, std::size_t precision, int align);
	void add_delimiter(char hchar, char xchar);
	std::string escape(const std::string& s);

	std::map<std::string, std::size_t> _column_indices;
	std::vector<boost::shared_ptr<Column> > _columns;
	TableFormat _format;
	std::size_t _width;
	std::size_t _rows;

}; // class


} // namespace

#endif /* FMG_TABLE_H_ */
