/*
 * Utils.cpp
 *
 *  Created on: 01.10.2012
 *      Author: Felix Ospald
 */

#include "Utils.h"

#include <string.h>
#include <stdio.h>
#include <stdarg.h>
#include <iomanip>

#include <dolfin/mesh/Vertex.h>
#include <dolfin/mesh/Edge.h>
#include <dolfin/mesh/Cell.h>

namespace fmg {

void Utils::random_vector(dolfin::GenericVector& x)
{
	uint n = x.size();
	for (uint i = 0; i < n; i ++) {
		x.setitem(i, rand() / (double) RAND_MAX);
	}
	x.apply("insert");
}


void Utils::print_system(const std::string& name, const dolfin::GenericMatrix& A, const dolfin::GenericVector& b)
{
	uint n = b.size();

	std::cout << name << std::endl;

	for (uint i = 0; i < n; i ++) {
		for (uint j = 0; j < n; j ++) {
			if (A(i,j) != 0.0) {
				std::cout
					<< std::fixed
					<< std::setprecision(1)
					<< std::setw(4)
					<< std::right
					<< (A(i,j)) << " ";
			}
			else {
				std::cout << "  0  ";
			}
		}
		std::cout << "| "
			<< std::fixed
			<< std::setprecision(4)
			<< std::setw(7)
			<< std::right
			<< (b[i]) << std::endl;
	}

	std::cout << std::endl;
}


void Utils::print_vector(const std::string& name, const dolfin::GenericVector& x)
{
	uint n = x.size();

	std::cout << name;

	for (uint i = 0; i < n; i ++) {
		std::cout
		/*	<< std::fixed
			<< std::setprecision(4)
			<< std::setw(7)
			<< std::right*/
			<< " " << (x[i]);
	}

	std::cout << std::endl;
}


void Utils::print_vector(const std::string& name, const std::vector<double>& x)
{
	uint n = x.size();

	std::cout << name;

	for (uint i = 0; i < n; i ++) {
		std::cout
			/*<< std::fixed
			<< std::setprecision(4)
			<< std::setw(7)
			<< std::right*/
			<< " " << (x[i]);
	}

	std::cout << std::endl;
}


void Utils::print_matrix(const std::string& name, const dolfin::GenericMatrix& A)
{
	uint n = A.size(0);
	uint m = A.size(1);

	std::cout << name << std::endl;

	for (uint i = 0; i < n; i ++) {
		for (uint j = 0; j < m; j ++) {
			if (A(i,j) != 0.0) {
				std::cout
					<< std::fixed
					<< std::setprecision(1)
					<< std::setw(4)
					<< std::right
					<< (A(i,j)) << " ";
			}
			else {
				std::cout << "  0  ";
			}
		}
		std::cout << std::endl;
	}

	std::cout << std::endl;
}


void Utils::print_mesh_entities(const dolfin::MeshEntity& entity, dolfin::uint dim)
{
	dolfin::uint n = entity.num_entities(dim);
	const dolfin::uint* e = entity.entities(dim);

	std::cout << "Entities(dim=" << dim << "):";

	for (dolfin::uint i = 0; i < n; i++) {
		std::cout << " " << e[i];
	}

	std::cout << std::endl;
}


void Utils::print_mesh_entity(const dolfin::MeshEntity& entity)
{
	dolfin::uint topologicalDim = entity.dim();

	for (dolfin::uint dim = 0; dim < topologicalDim; dim++) {
		print_mesh_entities(entity, dim);
	}

}


void Utils::print_mesh(const dolfin::Mesh& mesh)
{
	mesh.init();

	for (dolfin::VertexIterator v(mesh); !v.end(); ++v) {
		std::cout << "Vertex" << v->index() << ": " << v->point().str() << std::endl;
		print_mesh_entity(*v);
	}

	for (dolfin::EdgeIterator e(mesh); !e.end(); ++e){
		std::cout << "Edge" << e->index() << ": " << std::endl;
		print_mesh_entity(*e);
	}

	for (dolfin::CellIterator c(mesh); !c.end(); ++c)
	{
		std::cout << "Cell" << c->index() << ": " << std::endl;
		print_mesh_entity(*c);
	}

	std::cout << std::endl;
}


bool Utils::is_symmetric(const dolfin::GenericMatrix& A, double eps)
{
	return is_transpose(A, A, eps);
}


bool Utils::is_transpose(const dolfin::GenericMatrix& A, const dolfin::GenericMatrix& B, double eps)
{
	uint mA = A.size(0);
	uint nA = A.size(1);
	uint mB = B.size(0);
	uint nB = B.size(1);

	if (mA != nB) return false;
	if (nA != mB) return false;

	for (uint i = 0; i < mA; i ++) {
		for (uint j = 0; j < nA; j ++) {
			if (fabs(A(i,j) - B(j,i)) > eps) {
				return false;
			}
		}
	}

	return true;
}


void Utils::extract_subspaces_recursive(
		const boost::shared_ptr<const dolfin::FunctionSpace>& space,
		std::vector<boost::shared_ptr<const dolfin::FunctionSpace> >& list)
{
	dolfin::uint nsub = space->element()->num_sub_elements();

	if (nsub == 0) {
		list.push_back(space);
		return;
	}

	std::vector<dolfin::uint> component(1);

	for (dolfin::uint isub = 0; isub < nsub; isub++) {
		component[0] = isub;
		extract_subspaces_recursive(space->extract_sub_space(component), list);
	}
}


void Utils::split_string(std::vector<std::string>& vec, const std::string& s, const std::string& delim)
{
	// eqivalent to boost::algorithm::split_regex(vec, s, boost::regex("[delim]+"));

	const char* c = s.c_str();
	const char* start = c;
	const char* d = delim.c_str();

	vec.clear();

	for(;;)
	{
		if ((strchr(d, *c) != NULL) || (*c == '\0')) {
			if (start < c) {
				vec.push_back(std::string(start, (c-start)/sizeof(char)));
			}
			start = c+1;
		}

		if (*c == '\0') break;
		c++;
	}
}


std::vector<double>& Utils::temp(uint n)
{
	static std::vector<double> _temp;
	if (_temp.size() < n) _temp.resize(n, 0.0);
	return _temp;
}


const std::vector<double>& Utils::zero(uint n)
{
	static std::vector<double> _zero;
	if (_zero.size() < n) _zero.resize(n, 0.0);
	return _zero;
}


const std::vector<uint>& Utils::ints(uint n)
{
	static std::vector<uint> _ints;

	if (_ints.size() < n) {
		uint old_size = _ints.size();
		_ints.resize(n);
		for (uint i = old_size; i < n; i++) _ints[i] = i;
	}

	return _ints;
}


bool Utils::DEBUG = false;

void Utils::debug_print(const std::string& fmt, ...)
{
	char buffer[512];
	va_list aptr;
	va_start(aptr, fmt);
	std::string fmg_fmt("fmg: " + fmt);
	vsnprintf(buffer, sizeof(buffer), fmg_fmt.c_str(), aptr);
	va_end(aptr);
	//dolfin::log(dolfin::INFO, buffer);
	std::cout << buffer << std::endl;
}


} // namespace
