/*
 * Utils.h
 *
 *  Created on: 01.10.2012
 *      Author: Felix Ospald
 */

#ifndef FMG_UTILS_H_
#define FMG_UTILS_H_

#include <dolfin/la/GenericMatrix.h>
#include <dolfin/la/GenericVector.h>
#include <dolfin/mesh/Mesh.h>
#include <dolfin/mesh/MeshEntity.h>
#include <dolfin/function/FunctionSpace.h>

namespace fmg {

#define DEBUG_PRINT if (Utils::DEBUG) Utils::debug_print

/// This class contains some methods commonly used by fmg.

class Utils
{
public:
	/// Fill vector with [0,1] uniformly distributed random values.
	static void random_vector(dolfin::GenericVector& x);

	/// Print a vector.
	static void print_vector(const std::string& name, const dolfin::GenericVector& x);

	/// Print a vector.
	static void print_vector(const std::string& name, const std::vector<double>& x);

	/// Print a matrix.
	static void print_matrix(const std::string& name, const dolfin::GenericMatrix& A);

	/// Print a matrix and vector together.
	static void print_system(const std::string& name, const dolfin::GenericMatrix& A, const dolfin::GenericVector& b);

	/// Print mesh entities.
	static void print_mesh_entities(const dolfin::MeshEntity& entity, dolfin::uint dim);

	/// Print mesh entity.
	static void print_mesh_entity(const dolfin::MeshEntity& entity);

	/// Print mesh (vertices, edges, cells).
	static void print_mesh(const dolfin::Mesh& mesh);

	/// Check if matrix is symmetric.
	static bool is_symmetric(const dolfin::GenericMatrix& A, double eps = 1e-16);

	/// Check if matrices are transpose to each other.
	static bool is_transpose(const dolfin::GenericMatrix& A, const dolfin::GenericMatrix& B, double eps = 1e-16);

	/// Extract all subspaces of the given *space* and add them to *list*.
	static void extract_subspaces_recursive(
			const boost::shared_ptr<const dolfin::FunctionSpace>& space,
			std::vector<boost::shared_ptr<const dolfin::FunctionSpace> >& list);

	/// Split a string.
	static void split_string(std::vector<std::string>& vec, const std::string& s, const std::string& delim = ",; ");

	/// Returns a temporary vector of size >= n.
	static std::vector<double>& temp(uint n);

	/// Returns a zero vector of size >= n.
	static const std::vector<double>& zero(uint n);

	/// Returns the vector of integers 0 to size >= n.
	static const std::vector<uint>& ints(uint n);

	/// Print some debug text.
	static void debug_print(const std::string& fmt, ...);

	/// Switch to enable debugging mode of fmg.
	static bool DEBUG;

}; // class

} // namespace

#endif /* FMG_UTILS_H_ */
