/*
 * VectorCyclePattern.cpp
 *
 *  Created on: 02.10.2012
 *      Author: Felix Ospald
 */

#include "VectorCyclePattern.h"

#include <dolfin/log/log.h>
#include <iostream>

namespace fmg {


VectorCyclePattern::VectorCyclePattern(const std::vector<dolfin::uint>& cycles_per_level)
{
	reset();
	add_pattern(cycles_per_level);
	close_pattern();
}


VectorCyclePattern::VectorCyclePattern(const std::vector<std::vector<dolfin::uint> >& vec_cycles_per_level)
{
	reset();
	for (std::vector<std::vector<dolfin::uint> >::const_iterator i = vec_cycles_per_level.begin(); i != vec_cycles_per_level.end(); i++)
		add_pattern(*i);
	close_pattern();
}


VectorCyclePattern::VectorCyclePattern(const std::string& pattern, dolfin::uint number_of_levels)
{
	if (number_of_levels < 1) {
		dolfin::error("number of levels must be greater one!");
	}

	reset();

	std::vector<dolfin::uint> cycles_per_level;
	const char* p = pattern.c_str();
	dolfin::uint cycles = 0;

	do {
		char c = *p;

		if (isdigit(c)) {
			int n = (c - '0');
			cycles = (cycles > 0) ? (cycles*10+n) : n;
			if (cycles == 0)
				dolfin::error("cycle pattern must not contain zero cycles");
			if (cycles < 0)
				dolfin::error("overflow");
			continue;
		}

		if (cycles > 0) {
			// add cycle number
			if (cycles_per_level.size() >= number_of_levels)
				dolfin::error("cycle pattern exceeds number of levels");
			cycles_per_level.push_back(cycles);
			cycles = 0;
		}

		if (c == '*' || c == 'V') {
			// pad + concatenate
			cycles_per_level.resize(number_of_levels, 1);
			c = '-';
		}

		if (c == '-' || c == '\0') {
			// concatenate
			if (cycles_per_level.size() > 0) {
				add_pattern(cycles_per_level);
				cycles_per_level.clear();
			}
		}
	}
	while (*(p++));

	close_pattern();
}


void VectorCyclePattern::add_pattern(const std::vector<dolfin::uint>& cycles_per_level)
{
	// determine level offset if appending to existing pattern
	dolfin::uint level_offset = 0;

	if (_pattern.size() > 0) {
		level_offset = _pattern[_pattern.size()-1].level;
		switch (_pattern[_pattern.size()-1].action) {
			case Actions::prolongate: level_offset--; break;
			case Actions::restrict: level_offset++; break;
			case Actions::solve:
			case Actions::pr_transition:
				break;
		}
	}

	add_pattern(cycles_per_level, 0, level_offset);
}


void VectorCyclePattern::add_pattern(const std::vector<dolfin::uint>& cycles_per_level, dolfin::uint level, dolfin::uint level_offset)
{
	dolfin::uint cycles = (level < cycles_per_level.size()) ? cycles_per_level[level] : 1;

	if (cycles == 0)
		dolfin::error("cycle pattern must not contain zero cycles");

	PatternInfo pi;

	pi.level = level + level_offset;

	for (dolfin::uint i = 0; i < cycles; i++)
	{
		// solve on coarsest level
		if (level+1 == cycles_per_level.size()) {
			if (cycles > 1)
				dolfin::error("it does not make sense to do more the one cycle on the solve level");
			pi.action = Actions::solve;
			_pattern.push_back(pi);
			return;
		}

		// check for prolongate-restrict transition
		if ((_pattern.size() > 0) && (_pattern[_pattern.size()-1].action == Actions::prolongate)) {
			pi.action = Actions::pr_transition;
			_pattern.push_back(pi);
		}

		// restrict
		pi.action = Actions::restrict;
		_pattern.push_back(pi);
		pi.level++;

		// do recursive multigrid!
		add_pattern(cycles_per_level, pi.level, level_offset);

		// prolongate
		pi.action = Actions::prolongate;
		_pattern.push_back(pi);
		pi.level--;
	}
}


void VectorCyclePattern::close_pattern()
{
	if (_pattern.size() == 0) return;

	PatternInfo pi;

	// add missing prolongations to level 0
	while (_pattern[_pattern.size()-1].level-1 > _pattern[0].level) {
		pi.level = _pattern[_pattern.size()-1].level-1;
		pi.action = Actions::prolongate;
		_pattern.push_back(pi);
	}

	// add missing restrictions to level 0
	while (_pattern[_pattern.size()-1].level+1 < _pattern[0].level) {
		pi.level = _pattern[_pattern.size()-1].level-1;
		pi.action = Actions::restrict;
		_pattern.push_back(pi);
	}

	// check for prolongate-restrict transition when cycling to first action
	if ((_pattern[_pattern.size()-1].action == Actions::prolongate) && (_pattern[0].action == Actions::restrict)) {
		pi.level = _pattern[0].level;
		pi.action = Actions::pr_transition;
		_pattern.push_back(pi);
	}

#if 0
	for (dolfin::uint i = 0; i < _pattern.size(); i++) {
		std::cout << _pattern[i].level << "+" << _pattern[i].action << " ";
	}
	std::cout << "\n";
#endif
}


} // namespace
