/*
 * JacobiSmoothingOperator.cpp
 *
 *  Created on: 02.10.2012
 *      Author: Felix Ospald
 */

#include "JacobiSmoothingOperator.h"
#include "../common/common.h"

#if FMG_DOLFIN_VERSION > 1
	#include <dolfin/la/GenericLinearAlgebraFactory.h>
#else
	#include <dolfin/la/LinearAlgebraFactory.h>
#endif

namespace fmg {

JacobiSmoothingOperator::JacobiSmoothingOperator(
		const boost::shared_ptr<const dolfin::GenericMatrix>& A, double omega)
{
	if ((omega <= 0.0) || (omega >= 1.0))
		dolfin::warning("Jacobi relaxation parameter out of range (0,1)!");

	dolfin::GenericLinearAlgebraFactory& factory = A->factory();
	uint n = A->size(0);

	_A = A;
	_diag = FMG_FACTORY_VECTOR(factory);
	_diag->resize(n);
	_y = FMG_FACTORY_VECTOR(factory);
	_y->resize(n);

	// compute inverse of diagonal and scale with omega
	for (uint i = 0; i < n; i++) {
		_diag->setitem(i, omega / (*A)(i, i));
	}
}


void JacobiSmoothingOperator::apply(dolfin::GenericVector& x, const dolfin::GenericVector& b, bool& x_assumed_zero, uint reps) const
{
	// compute x = x - _diag * (_A*x - _b)

	for (uint i = 0; i < reps; i++)
	{
		if (x_assumed_zero) {
			x = b;
			x *= (*_diag);
			x_assumed_zero = false;
			continue;
		}

		_A->mult(x, *_y);
		(*_y) -= b;
		(*_y) *= (*_diag);
		x -= (*_y);
	}
}

} // namespace

