/*
 * SORSmoothingOperator.cpp
 *
 *  Created on: 25.10.2012
 *      Author: Felix Ospald
 */

#include "SORSmoothingOperator.h"
#include "../common/common.h"

#include <dolfin/la/PETScFactory.h>

#include <petscversion.h>
#if PETSC_VERSION_MAJOR==3 && PETSC_VERSION_MINOR>2
	#include <petsc-private/matimpl.h>
#else
	#include <private/matimpl.h>
#endif

namespace fmg {

typedef PetscErrorCode (*MatSOR_func_ptr)(Mat,Vec,PetscReal,MatSORType,PetscReal,PetscInt,PetscInt,Vec);
static MatSOR_func_ptr MatSOR_func = NULL;

SORSmoothingOperator::SORSmoothingOperator(const boost::shared_ptr<const dolfin::GenericMatrix>& A, SORType type, double omega)
{
	if ((omega <= 0.0) || (omega >= 2.0))
		dolfin::warning("SOR relaxation parameter out of range (0,2)!");

	switch (type) {
		case SORTypes::forward: _type = SOR_FORWARD_SWEEP; break;
		case SORTypes::backward: _type = SOR_BACKWARD_SWEEP; break;
		case SORTypes::symmetric: _type = SOR_SYMMETRIC_SWEEP; break;
		default:
			dolfin::dolfin_error(__FILE__, "constructor", "Type not supported.");
	}

	_A = A;
	_omega = omega;

	// this is a nasty hack to get around a bug in MatSOR_SeqAIJ_Inode
	// therefore we change the sor operation to the compatible MatSOR_SeqAIJ function
	// this is done by first creating a 0x0 SeqAIJ (without inodes) matrix and saving the function pointer
	// before we call MatSOR we exchange the pointers and then MatSOR will call a different implementation
	if (MatSOR_func == NULL)
	{
		PetscBool mat_no_inode = PETSC_FALSE, set;
		PetscOptionsGetBool(PETSC_NULL, "-mat_no_inode", &mat_no_inode, &set);

		if (!mat_no_inode) PetscOptionsSetValue("-mat_no_inode", "1");
		Mat* pX = new Mat();
		dolfin::PETScMatrix X(boost::shared_ptr<Mat>(pX, dolfin::PETScMatrixDeleter()));
		MatCreateSeqAIJ(PETSC_COMM_SELF, 0, 0, 0, PETSC_NULL, pX);
		MatSOR_func = (*pX)->ops->sor;
		if (!mat_no_inode) PetscOptionsSetValue("-mat_no_inode", "0");
	}

	if (MatSOR_func == NULL) {
		// fallback
		MatSOR_func = MatSOR;
		dolfin::warning("Please add command line argument --petsc.mat_no_inode to this program or "
				"dolfin::parameters.parse if you encounter problems with SOR smoothing");
	}
}


void SORSmoothingOperator::apply(dolfin::GenericVector& x, const dolfin::GenericVector& b, bool& x_assumed_zero, uint reps) const
{
	if (reps == 0) return;

	int flag = _type | (x_assumed_zero ? SOR_ZERO_INITIAL_GUESS : 0);

	//Utils::print_matrix("A: ", *_A);
	//Utils::print_vector("b: ", b);
	//Utils::print_vector("x_k: ", x);

	MatSOR_func(*(((dolfin::PETScMatrix*)_A.get())->mat()), *(((dolfin::PETScVector*)&b)->vec()),
		(PetscReal)_omega, (MatSORType)flag, (PetscReal)0.0, reps, reps,
		*(((dolfin::PETScVector*)&x)->vec()));

	x_assumed_zero = false;

	//Utils::print_vector("x_k+1: ", x);
}

} // namespace

