/*
 * MultigridPreconditionedKrylovSolver.cpp
 *
 *  Created on: 01.10.2012
 *      Author: Felix Ospald
 */

#include "MultigridPreconditionedKrylovSolver.h"
#include "../common/Utils.h"

namespace fmg {

MultigridPreconditionedKrylovSolver::MultigridPreconditionedKrylovSolver(MultigridSolver& mg_solver, const std::string& method) :
	dolfin::PETScKrylovSolver(decide_method(method, mg_solver.problem().is_symmetric()), *this),
	MultigridPreconditioner(mg_solver)
{
	init(dolfin::empty_parameters);
}


MultigridPreconditionedKrylovSolver::MultigridPreconditionedKrylovSolver(MultigridProblem& problem, const dolfin::Parameters& mg_params, const std::string& method) :
	dolfin::PETScKrylovSolver(decide_method(method, problem.is_symmetric()), *this),
	MultigridPreconditioner(boost::shared_ptr<MultigridSolver>(new MultigridSolver(problem)))
{
	init(mg_params);
}


MultigridPreconditionedKrylovSolver::MultigridPreconditionedKrylovSolver(dolfin::LinearVariationalProblem& problem,
	uint nrefine, bool symmetric, bool uniform, bool interpolate_coef, bool interpolate_solution,
	const dolfin::Parameters& mg_params, const std::string& method) :
	dolfin::PETScKrylovSolver(decide_method(method, symmetric), *this),
	MultigridPreconditioner(boost::shared_ptr<MultigridSolver>(new MultigridSolver(problem, nrefine, symmetric, uniform, interpolate_coef, interpolate_solution)))
{
	init(mg_params);
}


MultigridPreconditionedKrylovSolver::MultigridPreconditionedKrylovSolver(dolfin::NonlinearVariationalProblem& problem,
	uint nrefine, bool symmetric, bool uniform, bool interpolate_coef, bool interpolate_solution,
	const dolfin::Parameters& mg_params, const std::string& method) :
	dolfin::PETScKrylovSolver(decide_method(method, symmetric), *this),
	MultigridPreconditioner(boost::shared_ptr<MultigridSolver>(new MultigridSolver(problem, nrefine, symmetric, uniform, interpolate_coef, interpolate_solution)))
{
	init(mg_params);
}


void MultigridPreconditionedKrylovSolver::init(const dolfin::Parameters& mg_params)
{
	_mg_solver->parameters.update(mg_params);
	MultigridPreconditioner::init(_mg_solver);
	setup(*this);
}


std::string MultigridPreconditionedKrylovSolver::decide_method(const std::string& method, bool symmetric)
{
	if (method != "default")
		return method;
	return symmetric ? "cg" : "gmres";
}


uint MultigridPreconditionedKrylovSolver::solve(bool x_assumed_zero)
{
	if (!_mg_solver->initialized()) _mg_solver->init();

	dolfin::GenericVector& x = *(_mg_solver->solution()->vector());

	if (_mg_solver->reduce_required()) {
		if (!_xr) _xr = _mg_solver->create_reduced_vector();
		_mg_solver->reduce_vector(x, *_xr);
		uint r = solve(*_xr, x_assumed_zero);
		_mg_solver->expand_vector(*_xr, x);
		return r;
	}

	return solve(x, x_assumed_zero);
}


uint MultigridPreconditionedKrylovSolver::solve(dolfin::GenericVector& x, bool x_assumed_zero)
{
	if (!_mg_solver->initialized()) _mg_solver->init();

	return MultigridPreconditionedKrylovSolver::solve(x, *(_mg_solver->problem_rhs()), x_assumed_zero);
}


uint MultigridPreconditionedKrylovSolver::solve(dolfin::GenericVector& x, const dolfin::GenericVector& b, bool x_assumed_zero)
{
	if (!_mg_solver->initialized()) _mg_solver->init();

	// if we do not remove Dirichlet boundary conditions we need to start with x = b at the boundary dofs
	// this is not needed since the smoother does this automatically
	/*MultigridLevel& flevel = _mg_solver->finest_level();
	if (flevel.dbc_method == MultigridLevel::DirichletProjectionMethods::identify) {
		uint n = flevel.dbc_dofs.size();
		if (x_assumed_zero) x.zero();
		x_assumed_zero = false;
		b.get_local(&(Utils::temp(n)[0]), n, &(flevel.dbc_dofs[0]));
		x.set(&(Utils::temp(n)[0]), n, &(flevel.dbc_dofs[0]));
	}*/

	std::stringstream info;
	info << "fmg: solving linear system with " << x.size() << " unknowns";
	dolfin::info(info.str());

	bool nonzero_initial_guess_org = parameters["nonzero_initial_guess"];
	parameters["nonzero_initial_guess"] = !x_assumed_zero;

	try {
		this->set_operator(_mg_solver->problem_operator());
		uint n = dolfin::PETScKrylovSolver::solve(x, b);
		parameters["nonzero_initial_guess"] = nonzero_initial_guess_org;
		return n;
	}
	catch (...) {
		parameters["nonzero_initial_guess"] = nonzero_initial_guess_org;
		throw;
	}

	return 0;
}

} // namespace

