/*
 * MultigridProblem.cpp
 *
 *  Created on: 01.10.2012
 *      Author: Felix Ospald
 */

#include "MultigridProblem.h"
#include "../adapt/Adaptor.h"

#include <dolfin/function/FunctionSpace.h>

namespace fmg {

MultigridProblem::MultigridProblem(dolfin::LinearVariationalProblem& problem, uint nrefine,
		bool symmetric, bool uniform, bool interpolate_coef, bool interpolate_solution)
{
	_is_linear = true;
	_is_symmetric = symmetric;
	_is_uniform = true;
	add_problems(problem, uniform);
	for (uint i = 0; i < nrefine; i++)
		adapt(interpolate_coef, interpolate_solution);
}


MultigridProblem::MultigridProblem(dolfin::NonlinearVariationalProblem& problem, uint nrefine,
		bool symmetric, bool uniform, bool interpolate_coef, bool interpolate_solution)
{
	_is_linear = false;
	_is_symmetric = symmetric;
	_is_uniform = true;
	add_problems(problem, uniform);
	for (uint i = 0; i < nrefine; i++)
		adapt(interpolate_coef, interpolate_solution);
}


void MultigridProblem::adapt(bool interpolate_coef, bool interpolate_solution)
{
	boost::shared_ptr<const dolfin::Mesh> mesh = finest_level().V->mesh();
	fmg::Adaptor::adapt(*mesh);
	adapt(mesh->child(), true, interpolate_coef, interpolate_solution);
}


void MultigridProblem::adapt(const dolfin::MeshFunction<bool>& cell_markers, bool interpolate_coef, bool interpolate_solution)
{
	boost::shared_ptr<const dolfin::Mesh> mesh = finest_level().V->mesh();
	fmg::Adaptor::adapt(*mesh, cell_markers);
	adapt(mesh->child(), false, interpolate_coef, interpolate_solution);
}


void MultigridProblem::adapt(const dolfin::Mesh& refined_mesh, bool uniform, bool interpolate_coef, bool interpolate_solution)
{
	MultigridLevel& fine_level = finest_level();

	if (_is_linear) {
		dolfin::LinearVariationalProblem problem(fine_level.a, fine_level.L, fine_level.u, fine_level.bcs);
		fmg::Adaptor::adapt(problem, refined_mesh, interpolate_coef, interpolate_solution);
		add_problem(problem.child(), uniform);
	}
	else {
		dolfin::NonlinearVariationalProblem problem(fine_level.a, fine_level.u, fine_level.bcs, fine_level.L);
		fmg::Adaptor::adapt(problem, refined_mesh, interpolate_coef, interpolate_solution);
		add_problem(problem.child(), uniform);
	}
}


void MultigridProblem::adapt(const boost::shared_ptr<const dolfin::Mesh> & refined_mesh, bool uniform, bool interpolate_coef, bool interpolate_solution)
{
	adapt(*refined_mesh, uniform, interpolate_coef, interpolate_solution);
}


MultigridLevel* MultigridProblem::create_level() const
{
	return new MultigridLevel();
}


template<typename T> void MultigridProblem::add_problem(T& problem, bool uniform)
{
	boost::shared_ptr<MultigridLevel> level(create_level());

	if (_is_linear) {
		dolfin::LinearVariationalProblem* lp = (dolfin::LinearVariationalProblem*) &problem;
		level->a = lp->bilinear_form();
		level->L = lp->linear_form();
	}
	else {
		dolfin::NonlinearVariationalProblem* nlp = (dolfin::NonlinearVariationalProblem*) &problem;
		level->a = nlp->jacobian_form();
		level->L = nlp->residual_form();
	}

	level->V = problem.test_space();
	level->bcs = problem.bcs();
	level->u = problem.solution();
	level->index = _levels.size();
	level->is_linear = _is_linear;
	level->is_uniform = uniform;
	level->dbc_method = MultigridLevel::DirichletProjectionMethods::identify;

	uint nlevels = _levels.size();

	if (nlevels > 0) {
		_levels[nlevels-1]->is_uniform = uniform;
		_levels[nlevels-1]->finer_level = level;
		level->coarser_level = _levels[nlevels-1];
	}

	_levels.push_back(level);
	_is_uniform = _is_uniform && uniform;
}


template<typename T> void MultigridProblem::add_problems(T& problem, bool uniform)
{
	T* child = &problem;

	while (child != NULL)
	{
		add_problem(*child, uniform);
		child = child->child_shared_ptr().get();
	}
}


} // namespace
